using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace NetworkTest
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class ParallelPong : Microsoft.Xna.Framework.Game
    {
        #region Fields

        const int SCREEN_WIDTH = 1067;
        const int SCREEN_HEIGHT = 600;

        const int MAX_GAMERS = 2;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont font;

        MenuManager mMenus;
        Menu mHostOrJoinMenu;
        Menu mSelectGameToJoinMenu;
        bool mGameStarted = false;
        bool mIsReady = false;
        bool mIsHost = false;
        World mWorld;

        string mOpponent;

        NetworkSession mNetworkSession;

        PacketWriter mPacketWriter = new PacketWriter();
        PacketReader mPacketReader = new PacketReader();


        #endregion

        #region Initialization

        public ParallelPong()
        {
            graphics = new GraphicsDeviceManager(this);

            graphics.PreferredBackBufferWidth = SCREEN_WIDTH;
            graphics.PreferredBackBufferHeight = SCREEN_HEIGHT;

            Content.RootDirectory = "Content";
            mMenus = new MenuManager();

            mWorld = new World(SCREEN_WIDTH, SCREEN_HEIGHT);
            Components.Add(new GamerServicesComponent(this));
        }


        /// <summary>
        /// Load your content.
        /// </summary>
        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);

            font = Content.Load<SpriteFont>("SpriteFont1");
            
            mHostOrJoinMenu = new Menu(font, "Host Game?", true);
            mHostOrJoinMenu.AddMenuItem("Host Game", () => { mHostOrJoinMenu.Enabled = false; CreateSession(); });
            mHostOrJoinMenu.AddMenuItem("Join Game", () => { mHostOrJoinMenu.Enabled = false; SetUpJoinMenu(); });
            mHostOrJoinMenu.AddMenuItem("Quit",  Exit);
            mSelectGameToJoinMenu = new Menu(font, "Select Host To Join", false);
            mMenus.AddMenu(mSelectGameToJoinMenu);
            mMenus.AddMenu(mHostOrJoinMenu);
            mWorld.LoadContent(Content);
        }

        private void SetUpJoinMenu()
        {
            mSelectGameToJoinMenu.Enabled = true;
            try
            {
                AvailableNetworkSessionCollection availableSessions =
                                            NetworkSession.Find(NetworkSessionType.SystemLink,
                                                                1, null);
                foreach (AvailableNetworkSession s in availableSessions)
                {
                    mSelectGameToJoinMenu.AddMenuItem("Hosted by:" + s.HostGamertag, () => { mNetworkSession = NetworkSession.Join(s); 
                                                                                        HookSessionEvents(); 
                                                                                        mSelectGameToJoinMenu.Enabled = false; });
                }
                mSelectGameToJoinMenu.AddMenuItem("Back to Previous Menu", () => { mSelectGameToJoinMenu.Enabled = false; 
                                                                              mHostOrJoinMenu.Enabled = true; 
                                                                              mSelectGameToJoinMenu.ClearMenuItems(); });
                mIsHost = false;
            }
            catch (Exception e)
            {
                mSelectGameToJoinMenu.AddMenuItem("Could Not Initialize Network", null);

                mSelectGameToJoinMenu.AddMenuItem("Back to Previous Menu", () =>
                {
                    mSelectGameToJoinMenu.Enabled = false;
                    mHostOrJoinMenu.Enabled = true;
                    mSelectGameToJoinMenu.ClearMenuItems();
                });
            }

        }


        #endregion

        #region Update

        /// <summary>
        /// Allows the game to run logic.
        /// </summary>
        protected override void Update(GameTime gameTime)
        {
            if (IsActive && Gamer.SignedInGamers.Count == 0)
            {
                // If there are no profiles signed in, we cannot proceed.
                // Show the Guide so the user can sign in.
                Guide.ShowSignIn(1, false);
            }

            // Don't want our menus and the guide menus to collide 
            //   (that is, we don't want our menus and the Live Guide to both
            //    handle the same keypress/buttonpress, so we disable our menus
            //    when the guide is visible)
            if (!Guide.IsVisible)
            {
                mMenus.Update(gameTime);
            }

            if (mNetworkSession != null && mNetworkSession.RemoteGamers.Count > 0 && !mGameStarted)
            {
                // Really, we'd probably want to buffer this input.  Just quick & diry for now ...
                if (!mIsReady && (Keyboard.GetState().IsKeyDown(Keys.X) || 
                      GamePad.GetState(PlayerIndex.One).Buttons.X == ButtonState.Pressed))
                {
                    mIsReady = true;
                    foreach (LocalNetworkGamer g in mNetworkSession.LocalGamers)
                    {
                        g.IsReady = true;
                    }
                }
                if (mNetworkSession.IsEveryoneReady && mNetworkSession.IsHost)
                {
                    mNetworkSession.StartGame();
                }
            }
            else if (mNetworkSession != null && mGameStarted)
            {
                // If we are in a network session, update it.
                UpdateMainLoop(gameTime);
            }

            if (mNetworkSession != null)
            {
                mNetworkSession.Update();
            }

            base.Update(gameTime);
        }


        /// <summary>
        /// Starts hosting a new network session.
        /// </summary>
        void CreateSession()
        {
            try
            {
                mNetworkSession = NetworkSession.Create(NetworkSessionType.SystemLink, 1, MAX_GAMERS);
                mIsHost = true;

                HookSessionEvents();
            }
            catch (Exception e)
            {

            }
        }


        /// <summary>
        /// After creating or joining a network session, we must subscribe to
        /// some events so we will be notified when the session changes state.
        /// </summary>
        void HookSessionEvents()
        {
            mNetworkSession.GamerJoined += GamerJoinedEventHandler;
            mNetworkSession.SessionEnded += SessionEndedEventHandler;
            mNetworkSession.GameStarted += GameStartedHandler;
        }


        void GameStartedHandler(object sender, GameStartedEventArgs e)
        {
            mWorld.InitializeWorld(mIsHost);
            mGameStarted = true;           
        }

        /// <summary>
        /// This event handler will be called whenever a new gamer joins the session.
        /// We use it to allocate a Tank object, and associate it with the new gamer.
        /// </summary>
        void GamerJoinedEventHandler(object sender, GamerJoinedEventArgs e)
        {
            // We get this callback when *we* join, as well as when anyone else joins.
            // Note that in a 2-player game, both players will get two callbacks
            // (one for themselves, and one for the other player)
            if (!e.Gamer.IsLocal)
            {
                mOpponent = e.Gamer.Gamertag;
            }
        }


        /// <summary>
        /// Event handler notifies us when the network session has ended.
        /// </summary>
        void SessionEndedEventHandler(object sender, NetworkSessionEndedEventArgs e)
        {
            mNetworkSession.Dispose();
            mNetworkSession = null;
        }


        /// <summary>
        /// Updates the state of the network session, moving the tanks
        /// around and synchronizing their state over the network.
        /// </summary>
        void UpdateMainLoop(GameTime gameTime)
        {
            HandlePaddleInput();
            mWorld.Update(gameTime);

            // Send out output to the world
            // Currently, we are only using one local gamer (no split screen / etc)
            // the system assumes that there could be more than one, so LocalGamers is
            // a collection.
            SendDataToNetwork(mNetworkSession.LocalGamers[0]);
            GetDataFromNetwork(mNetworkSession.LocalGamers[0]);
        }

        private void HandlePaddleInput()
        {
            KeyboardState ks = Keyboard.GetState();
            Vector2 stick = GamePad.GetState(PlayerIndex.One).ThumbSticks.Left;
            mWorld.mMyPaddle.mVelocity = Vector2.Zero;
            if (ks.IsKeyDown(Keys.Down) || stick.Y < -0.5f)
            {
                mWorld.mMyPaddle.mVelocity = new Vector2(0,1000);
            }
            if (ks.IsKeyDown(Keys.Up) || stick.Y > 0.5f)
            {
                mWorld.mMyPaddle.mVelocity = new Vector2(0, -1000);
            }
        }


        /// <summary>
        /// Helper for updating a locally controlled gamer.
        /// </summary>
        void SendDataToNetwork(LocalNetworkGamer gamer)
        {
            mPacketWriter.Write(mWorld.mMyPaddle.mObjectSprite.Position);
            gamer.SendData(mPacketWriter, SendDataOptions.InOrder);
        }


        /// <summary>
        /// Helper for reading incoming network packets.
        /// </summary>
        void GetDataFromNetwork(LocalNetworkGamer gamer)
        {
            // Keep reading as long as incoming packets are available.
            while (gamer.IsDataAvailable)
            {
                NetworkGamer sender;

                // Read a single packet from the network.
                gamer.ReceiveData(mPacketReader, out sender);

                // Discard packets sent by local gamers: we already know their state!
                if (sender.IsLocal)
                    continue;


                Vector2 opponentPosition = mPacketReader.ReadVector2();
                // Read any other data we want to get from this player

                mWorld.mOpponentPaddle.mObjectSprite.Position = opponentPosition;
            }
        }


        #endregion

        #region Draw


        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);
            spriteBatch.Begin();
            mMenus.Draw(spriteBatch);

            if (mNetworkSession != null && mNetworkSession.RemoteGamers.Count == 0)
            {
                spriteBatch.DrawString(font, "Waiting for player to connect", new Vector2(20, 20), Color.White);
            }
            else if (mNetworkSession  != null && !mGameStarted)
            {
                spriteBatch.DrawString(font, "Waiting for players to be ready", new Vector2(20, 20), Color.White);

                if (!mIsReady)
                {
                    spriteBatch.DrawString(font, "Press X to signal you are ready", new Vector2(20, 50), Color.White);
                }
                else
                {
                    spriteBatch.DrawString(font, "You are ready.  Waiting for other player ... ", new Vector2(20, 50), Color.White);
                }
            }

            if (mGameStarted)
            {
                mWorld.Draw(spriteBatch);
            }

            spriteBatch.End();

            base.Draw(gameTime);
        }

        #endregion

    }
}
